/*
   Small library of useful utilities based on the dietlib by fefe
   Copyright (C) 2003, 2008 Andreas Franz Borchert
   --------------------------------------------------------------------
   This library is free software; you can redistribute it and/or modify
   it under the terms of the GNU Library General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*

=head1 NAME

mpx_session -- run a multiplexed network service on a given port

=head1 SYNOPSIS

   #include <afblib/mpx_session.h>

   typedef struct session {
      void* handle;
      stralloc request;
      stralloc response;
   } session;

   typedef void (*request_handler)(session* s);
   typedef void (*hangup_handler)(session* s);

   void send_response(session* s);
   void close_session(session* s);
   void run_mpx_service(in_port_t port,
      request_handler rhandler, hangup_handler hhandler);

=head1 DESCRIPTION

I<run_mpx_service> creates a socket that listens on I<port>,
accepts all incoming connections, invokes I<rhandler> for
every newline-terminated input line from any of the network
connections, and calls I<hhandler> whenever a network connection
terminates.

Request handlers get a session reference with following fields:

=over 4

=item I<handle>

This field is initially null and may be freely used to maintain
a pointer to a custom session object.

=item I<request>

This I<stralloc> object contains one input line from the given
connection. The terminating line-feed, and a possibly preceding
carriage-return are already removed.

=item I<response>

This I<stralloc> object is initially empty whenever a request
handler gets called. This string object may be filled (without
terminating null-bytes, carriage-returns or line-feeds) to create
a response packet. If this field remains empty, no response is
sent back. Instead, a delayed reponse may be sent any time later
on by using I<send_response>.

=back

I<close_session> may be called to initiate the shutdown 
of a session, i.e. no further input packets will be processed,
just the pending response packets will be taken care of.

I<run_mpx_service> runs normally infinitely and returns in
error cases only.

=head1 EXAMPLE

Following example implements a network service that allows
to increment whole numbers. Usually, every session has its
own number that gets incremented. But, by adding the keyword
B<global> to an increment, a global counter that is shared
over all sessions gets incremented.

   #include <afblib/mpx_session.h>
   #include <afblib/tokenizer.h>
   #include <stralloc.h>
   #include <stdlib.h>
   #include <stdio.h>
   #include <string.h>

   struct status {
      int count; // per-session counter
   };
   int global_counter = 0; // global counter, shared by all sessions

   void handle_request(session* s) {
      if (s->handle == 0) {
	 // new network connection, create a new status object for it
	 struct status* sp = (struct status*) malloc(sizeof(struct status));
	 if (sp == 0) return;
	 sp->count = 0;
	 s->handle = (void*) sp;
      }
      struct status* sp = (struct status*) s->handle;
      strlist tokens = {0};
      stralloc_0(&s->request); // 0-termination required by tokenizer
      if (!tokenizer(&s->request, &tokens)) return;
      int value = 0;
      if (tokens.len > 0) {
	 int increment;
	 if (sscanf(tokens.list[0], "%d", &increment) == 1) {
	    if (tokens.len == 2 && strcmp(tokens.list[1], "global") == 0) {
	       global_counter += increment;
	       value = global_counter;
	    } else {
	       sp->count += increment; 
	       value = sp->count;
	    }
	 }
      }
      // return current value of the chosen counter as response
      stralloc_catint(&s->response, value);
   }

   void handle_close(session* s) {
      if (s->handle) {
	 free(s->handle);
      }
   }

   int main() {
      run_mpx_service(47110, handle_request, handle_close);
   }

=head1 AUTHOR

Andreas F. Borchert

=cut

*/

#include <afblib/mpx_session.h>
#include <afblib/multiplexor.h>
#include <netinet/in.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <time.h>
#include <unistd.h>

typedef struct mpx_service {
   request_handler rhandler;
   hangup_handler hhandler;
} mpx_service;

void send_response(session* s) {
   if (s->response.len > 0) {
      /* send response packet, if any */
      stralloc_cats(&s->response, "\r\n");
      stralloc_0(&s->response);
      char* buf = strdup(s->response.s);
      if (buf == 0) return;
      if (!write_to_link(s->link, buf, s->response.len-1)) return;
   }
   /* truncate both buffers */
   s->response.len = 0;
   s->request.len = 0;
}

void mpx_input_handler(connection* link, char* buf, unsigned int len) {
   if (link->handle == 0) {
      /* new connection */
      session* newsession = malloc(sizeof(session));
      if (newsession == 0) return;
      *newsession = (session) {0};
      link->handle = (void*) newsession;
      newsession->link = link;
   }
   session* s = (session*) link->handle;
   mpx_service* mpxs = (mpx_service*) link->mpx_handle;
   /* process input */
   for (int index = 0; index < len; ++index) {
      char ch = buf[index];
      if (ch == '\n') {
	 /* process completed request package */
	 if (s->request.len > 0 && s->request.s[s->request.len-1] == '\r') {
	    --(s->request.len);
	 }
	 s->response.len = 0;
	 (*(mpxs->rhandler))(s);
	 send_response(s);
      } else {
	 stralloc_append(&s->request, &ch);
      }
   }
   free(buf);
}

void mpx_close_handler(connection* link) {
   session* s = (session*) link->handle;
   mpx_service* mpxs = (mpx_service*) link->mpx_handle;
   if (s) {
      if (mpxs->hhandler) {
	 (*(mpxs->hhandler))(s);
      }
      stralloc_free(&s->request);
      stralloc_free(&s->response);
      free(s);
   }
}

void close_session(session* s) {
   close_link(s->link);
}

void run_mpx_service(in_port_t port,
      request_handler rhandler, hangup_handler hhandler) {
   mpx_service* mpxs = malloc(sizeof(mpx_service));
   if (mpxs == 0) return;
   *mpxs = (mpx_service) {rhandler, hhandler};
   struct sockaddr_in address = {0};
   address.sin_family = AF_INET;
   address.sin_addr.s_addr = htonl(INADDR_ANY);
   address.sin_port = htons(port);

   int sfd = socket(PF_INET, SOCK_STREAM, 0);
   int optval = 1;
   if (sfd < 0 ||
        setsockopt(sfd, SOL_SOCKET, SO_REUSEADDR,
                   &optval, sizeof optval) < 0 ||
        bind(sfd, (struct sockaddr *) &address,
             sizeof address) < 0 ||
        listen(sfd, SOMAXCONN) < 0) {
      return;
   }
   run_multiplexor(sfd, mpx_input_handler, mpx_close_handler, (void*) mpxs);
}
